let fs = require('fs');
let requestService = require('./request');
let paginationService = require('./pagination');
let responseService = require('./response');
let linksService = require('./links');
let AuthService = require('./auth');
let sort = require('./sort');
let atob = require('atob');

function DynamicRoutesService(app, host) {

    let authService = new AuthService(app);
    let metadataPath = __dirname + '/../metadata/database/';
    let files = fs.readdirSync(metadataPath);
    let fileContents = {}, data = {};
    let params = { apiUri: 'http://' + host };

    /**
     * Creates all metadata routes
     */
    this.createRoutes = () => {
        readMetadata();
        files.forEach((file) => {
            let filename = getFilename(file);
            let fileContent = fileContents[filename];
            let middleware = getAuthMiddleware(fileContent.requireAuth);
            let route = '/' + filename;
            createSingleRoutes(route, filename, middleware, fileContent);
            if (fileContent.relations) {
                createNestedRoutes(filename, middleware, fileContent);
            }
        });
    };

    /**
     * Reads all metadata files
     */
    let readMetadata = () => {
        files.forEach((file) => {
            fileContents[getFilename(file)] = getFileContent(metadataPath + file);
        });
    };

    /**
     * Creates all routes for each metadata file
     * @param {String} publicRoute Client request route
     * @param {String} filename Metadata filename
     * @param {Function} middleware Function to intercept the request
     * @param {Object} fileContent Metadata file content
     */
    let createSingleRoutes = (publicRoute, filename, middleware, fileContent) => {
        // get all route
        createRoute(publicRoute, filename, middleware, 'get', fileContent.methods.get, sendGetAllResponse);
        // get one route
        createRoute(publicRoute + '/:id', filename, middleware, 'get', fileContent.methods.get, sendGetOneResponse);
        // post route
        createRoute(publicRoute, filename, middleware, 'post', fileContent.methods.post, sendPostResponse);
        // put route
        createRoute(publicRoute + '/:id', filename, middleware, 'put', fileContent.methods.put, sendPutResponse);
        // delete route
        createRoute(publicRoute + '/:id', filename, middleware, 'delete', fileContent.methods.delete, sendDeleteResponse);
    };

    /**
     * Creates all nested routes
     * @param {String} parentRoute Nested route
     * @param {Function} middleware Function to intercept the request
     * @param {Object} parentContent Parent metadata file content
     */
    let createNestedRoutes = (parentRoute, middleware, parentContent) => {
        for (let relationKey in parentContent.relations) {
            let relation = parentContent.relations[relationKey];
            if (!relation.asFilter) {
                let paramName = relation.column;
                let route = `/${parentRoute}/:${paramName}/${relationKey}`;
                createRoute(route, relationKey, middleware, 'get', fileContents[relationKey].methods.get, sendGetAllResponse);
                createRoute(route + '/:id', relationKey, middleware, 'get', fileContents[relationKey].methods.get, sendGetOneResponse);
            }
        }
    };

    /**
     * Creates route based on metadata
     * @param {String} publicRoute Client request route
     * @param {String} route Metadata filename
     * @param {Function} middleware Function to intercept the request
     * @param {Boolean} allowedMethod
     * @param {Function} callback Response callback
     */
    let createRoute = (publicRoute, route, middleware, method, allowedMethod, callback) => {
        app[method](publicRoute, middleware, (req, res) => {
            if (!allowedMethod) {
                responseService.throwMethodNotAllowed(res);
            } else {
                createRouteResponse(req, res, route, callback);
            }
        });
    };

    /**
     * Checks if a record exists before execute requested action
     * @param {Reponse} res Node response
     * @param {String} reqId Row id
     * @param {Array} routeData Dataset
     * @param {boolean} exact exact key
     * @param {Function} successFunction Success callback
     */
    let checkRecordExists = (res, req, exact, routeData, successFunction) => {
        let params = Object.assign(extractQueryFilter(req.query), req.params);
        let indexFound = findRowIndex(routeData,exact ,params);
        if (indexFound !== -1) {
            successFunction(indexFound);
        } else {
            responseService.throwNotFound(res);
        }
    };

    /**
     * Sends all data to client
     * @param {Request} req Node request
     * @param {Response} res Node response
     * @param {String} route Metadata filename
     * @param {String} completeUri API address with route
     */
    let sendGetAllResponse = (req, res, route, completeUri) => {
        let page = req.query.page || 1;
        let limit = req.query.limit || data[route].length;
        let result = buildGetAllDataset(req, route, page, limit);
        res.json({
            data: linksService.getAllLinks(result.data, completeUri, fileContents[route].relations, params.apiUri),
            pagination: paginationService.getPagination(result.all, page, completeUri, limit),
        });
    };

    /**
     * Sends a specific data to client
     * @param {Request} req Node request
     * @param {Response} res Node response
     * @param {String} route Metadata filename
     * @param {String} completeUri API address with route
     */
    let sendGetOneResponse = (req, res, route, completeUri) => {
        let routeData = data[route];
        const exact = fileContents[route].exact
        checkRecordExists(res, req, exact, routeData, (indexFound) => {
            let row = formatRow(routeData[indexFound], fileContents[route].columns, 'showOnView');
            res.json({ data: linksService.getLinks(row, completeUri, fileContents[route].relations, params.apiUri) });
        });
    };

    /**
     * Creates a new record
     * @param {Request} req Node request
     * @param {Response} res Node response
     * @param {String} route Metadata filename
     * @param {String} completeUri API address with route
     */
    let sendPostResponse = (req, res, route, completeUri) => {
        let body = req.body;
        body.id = fileContents[route].nextId;
        data[route].push(body);
        fileContents[route].nextId += 1;
        res.json({ data: { id: body.id, uri: linksService.getPostLinks(body.id, completeUri) }});
    };

    /**
     * Updates a record
     * @param {Request} req Node request
     * @param {Response} res Node response
     * @param {String} route Metadata filename
     * @param {String} completeUri API address with route
     */
    let sendPutResponse = (req, res, route, completeUri) => {
        let routeData = data[route];
        const exact = fileContents[route].exact
        checkRecordExists(res, req, exact, routeData, (indexFound) => {
            routeData[indexFound] = Object.assign(routeData[indexFound], req.body);
            res.json({ data: linksService.getEditLinks(routeData[indexFound], completeUri) });
        });
    };

    /**
     * Deletes a record
     * @param {Request} req Node request
     * @param {Response} res Node response
     * @param {String} route Metadata filename
     * @param {String} completeUri API address with route
     */
    let sendDeleteResponse = (req, res, route) => {
        let routeData = data[route];
        const exact = fileContents[route].exact
        checkRecordExists(res, req, exact, routeData, (indexFound) => {
            let removedRow = routeData.splice(indexFound, 1)[0];
            res.json({ data: { id: removedRow.id }});
        });
    };

    /**
     * Retrieve a builded dataset filtered by query string and request params
     * @param {Request} req Node request
     * @param {String} route Metadata filename
     * @param {Number} page Dataset page number
     * @param {Number} limit Items per page
     * @returns {Array} Builded dataset
     */
    let buildGetAllDataset = (req, route, page, limit) => {
        let dataset = data[route];
        let end = limit  * page;
        let start = end - limit;

        if (req.query.dynamic_filter) {
            const dynamicFilter = JSON.parse(atob(req.query.dynamic_filter));
            dataset = dataset.filter(compareDynamicFilter, dynamicFilter);
            delete req.query.dynamic_filter;
        }

        if (Object.keys(req.params).length || Object.keys(req.query).length) {
            let params = Object.assign(extractQueryFilter(req.query), req.params);
            if (fileContents[route].exact) {
                dataset = dataset.filter(compareFilterRowExact, params);
            } else {
                dataset = dataset.filter(compareFilterRow, params);
            }
            if (!fileContents[route].simple) {
                dataset = req.query.search
                    ? dataset.filter(row => getRowBySearch(row, req.query.search_in, req.query.search))
                    : dataset;
            } else if (req.query.search) {
                dataset = req.query.search
                    ? dataset.filter(row => row.indexOf(req.query.search) !== -1)
                    : dataset;
            }
            orderResponse(req.query.order, dataset);
        }
        let formatted = dataset.slice(start, end);
        if (!fileContents[route].simple) {
            formatted = formatDataset(formatted, route, 'showOnList');
        }
        return { data: formatted, all: dataset};
    };

    let getRowBySearch = (row, searchIn, search) => {
        let filtered = false;
        for (const key in row) {
          if (searchIn.indexOf(key) !== -1 && row[key].toString().indexOf(search) !== -1) {
            filtered = true;
            break;
          } else {
            filtered = false;
          }
        }
        return filtered;
    }

    /**
     * Request data and execute route callback
     * @param {Request} req Node request
     * @param {Response} res Node response
     * @param {String} route Metadata filename
     * @param {Function} callback Function to execute after data is requested
     */
    let createRouteResponse = (req, res, route, callback) => {
        let completeUri = params.apiUri + '/' + route;
        if (fileContents[route] && fileContents[route].requested) {
            callback.call(null, req, res, route, completeUri);
        } else {
            requestRouteData(route).then(() => {
                fileContents[route].nextId = data[route].length + 1;
                callback.call(null, req, res, route, completeUri);
            }).catch((error) => {
                res.status(500).json({ error });
            });
        }
    };

    /**
     * Requests a data from api or create custom local data
     * @param {String} route Metadata filename
     * @returns {Promise.<void>} Empty promise
     */
    let requestRouteData = async (route) => {
        fileContents[route].requested = true;
        let body, content = fileContents[route];
        if (content.externalApi) {
            try {
                body = await requestService.get(content.apiUrl);
                data[route] = getBuildedBody(body, content);
            } catch (e) {
                fileContents[route].requested = false;
                throw e;
            }
        } else if (content.static) {
            data[route] = await Promise.resolve(content.data);
        } else {
            data[route] = await Promise.resolve(getLocalData(content));
        }
    };

    /**
     * Retrieves a local data based on metadata
     * @param {Object} content Metadata content
     * @returns {Array.<Object>}
     */
    let getLocalData = (content) => {
        let body = [], newRow;
        for (let i = 1; i <= content.count; i++) {
            if (!content.simple) {
                newRow = getLocalRow(content.columns, i);
            } else {
                newRow = getColumnValue(content.simpleConfig, i);
            }
            body.push(newRow);
        }
        return body;
    };

    /**
     * Retrieves a row based on metadata
     * @param {Object} columns Metadata columns
     * @param {Number} index Current key
     * @returns {Object}
     */
    let getLocalRow = (columns, index) => {
        let newRow = {};
        for (var key in columns) {
            if (columns[key].foreignKey) {
                newRow[key] = getRandomDecimal(1, fileContents[columns[key].table].count, 0);
            } else {
                newRow[key] = getColumnValue(columns[key], index);
            }
        }
        return newRow;
    };

    /**
     * Retrieves a value for a column
     * @param {Object} column Metadata column
     * @param {Number} index Current key
     * @returns {Number|String}
     */
    let getColumnValue = (column, index) => {
        switch (column.type) {
            case 'int':
                return +(column.value.replace(/\%s/g, index));
            case 'string':
                return column.value.replace(/\%s/g, index);
            case 'float':
                return getRandomDecimal(column.min, column.max, column.precision);
            case 'date':
                return randomDate(new Date(2012, 0, 1), new Date())
        }
        return 'type not supported';
    };

    let randomDate = (start, end) => {
        let date = new Date(start.getTime() + Math.random() * (end.getTime() - start.getTime()));
        return date.getFullYear() + '-' + (date.getMonth() + 1).toString().padStart(2, '0') + '-' + date.getDate().toString().padStart(2, '0');
    }

    /**
     * Retrieves an array of builded body
     * @param {*} body Request body
     * @param {Object} content Metadata content
     * @returns {Array.<Object>}
     */
    let getBuildedBody = (body, content) => {
        let keys = Object.keys(content.columns);
        let bodyParsed = getBodyParsed(body, content.dataColumn);
        return getFormattedBody(bodyParsed, content.columns, keys);
    };

    /**
     * Retrieves an array of formatted body
     * @param {Array.<Object>} bodyParsed Request body only with necessary values
     * @param {Object} columns Metadata columns
     * @param {Array.<String>} keys Column keys
     * @returns {Array.<Object>}
     */
    let getFormattedBody = (bodyParsed, columns, keys) => {
        let formattedBody = [], newRow;
        bodyParsed.forEach((row) => {
            newRow = {};
            keys.forEach((key) => {
                newRow[key] = getNestedObject(row, columns[key].location);
            });
            formattedBody.push(newRow);
        });
        return formattedBody;
    };

    /**
     * Parses and retrieves the body requested
     * @param {String} body Requested result
     * @param {String} column Result column where data is located
     * @returns {Object}
     */
    let getBodyParsed = (body, column) => {
        let bodyParsed = JSON.parse(body);
        if (column) {
            bodyParsed = bodyParsed[column];
        }
        return bodyParsed;
    };

    /**
     * Retrieves a row index
     * @param {Array} data Dataset
     * @param {Object} params Filter params
     * @param {boolean} exact exact key
     * @returns {Number} Index found
     */
    let findRowIndex = (data, exact, params) => {
        return data.findIndex(exact ? compareFilterRowExact : compareFilterRow, params);
    };

    /**
     * Retrieves a function to intercept the request
     * @param {Boolean} requireAuth Metadata require auth param
     * @returns {Function}
     */
    let getAuthMiddleware = (requireAuth) => {
        return requireAuth ? authService.isAutheticated : (req, res, next) => { return next(); };
    };

    /**
     * Retrieves a file content
     * @param {String} path Metadata file path
     * @returns {*} File content
     */
    let getFileContent = (path = '') => {
        return JSON.parse(fs.readFileSync(path, 'utf8'));
    };

    /**
     * Retrieves a filename without extension
     * @param {String} fileWithExtension Filename with extension
     * @returns {String} Filename
     */
    let getFilename = (fileWithExtension = '') => {
        return fileWithExtension.split('.')[0];
    };

    /**
     * Retrieves any nested property from an object
     * @param {Object} [nestedObj={}] Any object
     * @param {Array.<String>} [pathArr=[]] An arrays of strings
     * @returns {*}
     */
    let getNestedObject = (nestedObj = {}, pathArr = []) => {
        return pathArr.reduce((obj, key) =>
            (obj && obj[key] !== 'undefined') ? obj[key] : undefined, nestedObj);
    };

    /**
     * Retrieves a random decimal number from specific range
     * @param {Number} [min=1] Interger number.
     * @param {Number} [max=1000] Interger number.
     * @param {Number} [precision=2] Interger number.
     * @returns {Number} Float number.
     */
    let getRandomDecimal = (min = 1, max = 1000, precision = 2) => {
        precision = Math.pow(10, precision);
        return Math.round(getRandomNumber(min, max) * precision) / precision;
    };

    /**
     * Retrieves a random number from specific range.
     * @param {Number} [min=0] Interger number.
     * @param {Number} [max=1000] Interger number.
     * @returns {Number} Float number
     */
    let getRandomNumber = (min = 0, max = 1000) => {
        return Math.random() * (max - min) + min;
    };

    /**
     * Retrieves a dataset with configured columns
     * @param {Array} dataset Metadata dataset
     * @param {String} route Metadata filename
     * @param {String} visibility Route visibility type
     * @returns {Array} Formatted dataset
     */
    let formatDataset = (dataset, route, visibility) => {
        if (fileContents[route].static) {
            return dataset;
        }

        let formattedDataset = [], columns = fileContents[route].columns;
        dataset.forEach((row) => {
            formattedDataset.push(formatRow(row, columns, visibility));
        });
        return formattedDataset;
    };

    /**
     * Retrieves a row with configured columns
     * @param {Object} row Dataset row
     * @param {Object} columns Metadata columns
     * @param {String} visibility Route visibility type
     * @returns {Object} Formatted row
     */
    let formatRow = (row, columns, visibility) => {
        let newRow = {}
        for (let column in columns) {
            if (columns[column][visibility] !== false) {
                newRow[column] = row[column];
            }
        }
        return newRow;
    };

    /**
     * Retrieves a boolean for row comparison
     * @param {Object} row Dataset row
     * @returns {Boolean}
     */
    function compareFilterRow(row) {
        if (!Object.keys(this).length) {
            return true;
        }
        let filtered = false;
        for (var key in this) {
            const values = Array.isArray(this[key]) ? this[key] : [this[key]];

            filtered = values.some((value) => !row[key] || row[key].toString().indexOf(value) !== -1);

            if (!filtered) {
                break;
            }
        }
        return filtered;
    };

    /**
     * Retrieves a boolean for row exact comparison
     * @param {Object} row Dataset row
     * @returns {Boolean}
     */
    function compareFilterRowExact(row) {
        if (!Object.keys(this).length) {
            return true;
        }
        let filtered = false;
        for (var key in this) {
            if (this[key] === 'NULL') {
                filtered = (row[key] == null);
            } else {
                filtered = row[key] && row[key].toString() === this[key].toString();
            }

            if (!filtered) {
                break;
            }
        }
        return filtered;
    };

    /**
     * Retrieves a boolean for row dynamic filter comparison
     * @param {Object} row Dataset row
     * @returns {Boolean}
     */
    function compareDynamicFilter(row) {
        if (!Object.keys(this).length) {
            return true;
        }
        let filtered = undefined;
        for (var key in this) {
            const filterItems = this[key];
            filterItems.forEach((item) => {
                if (filtered === false && item.relation === 'AND') return;
                if (filtered === true && item.relation === 'OR') return;

                const columnValue = row[key].toString();
                const value = item.value.toString();

                switch (item.operation) {
                    case 'CONTAINS':
                        filtered = columnValue.indexOf(value) !== -1;
                        break;
                    case 'NOT_CONTAINS':
                        filtered = columnValue.indexOf(value) === -1;
                        break;
                    case 'EQUALS':
                        filtered = columnValue === value;
                        break;
                    case 'NOT_EQUALS':
                        filtered = columnValue !== value;
                        break;
                    case 'GREATER_THAN':
                        filtered = (Number(columnValue) || columnValue) > (Number(value) || value);
                        break;
                    case 'LESS_THAN':
                        filtered = (Number(columnValue) || columnValue) < (Number(value) || value);
                        break;
                    case 'GREATER_THAN_EQUALS':
                        filtered = (Number(columnValue) || columnValue) >= (Number(value) || value);
                        break;
                    case 'LESS_THAN_EQUALS':
                        filtered = (Number(columnValue) || columnValue) <= (Number(value) || value);
                        break;
                    case 'IN':
                        filtered = value.split(';').includes(columnValue);
                        break;
                    case 'NOT_IN':
                        filtered = !value.split(';').includes(columnValue);
                        break;
                    case 'BETWEEN':
                        const range = value.split(';');
                        filtered = (Number(columnValue) || columnValue) >= (Number(range[0]) || range[0])
                            && (Number(columnValue) || columnValue) <= (Number(range[1]) || range[1]);
                        break;
                    default:
                        break;
                }

            });
        }
        return filtered;
    };

    /**
     * Retrieves a query filter object
     * @param {Object} query Request query params
     * @returns {Object}
     */
    let extractQueryFilter = (query) => {
        let queryFilter = {};
        if (query.q)  {
            queryFilter = Object.assign(queryFilter, JSON.parse(query.q));
        }
        for (let key in query) {
            if (key != "page" && key != "limit" && key != "order" && key != "filter" && key != "search" && key != "in" && key != "search_in") {
                queryFilter[key] = query[key];
            }
        }
        return queryFilter;
    };

    let orderResponse = (queryOrder, dataset) => {
        if (queryOrder) {
            let columns = [], order = [];
            if (typeof queryOrder === 'string') {
                queryOrder = [queryOrder];
            }
            queryOrder.forEach((value) => {
                const splitted = value.split('.');
                columns.push(splitted[0]);
                order.push(splitted[1]);
            });
            sort.sortArray(dataset, columns, order);
        }
    }
}

module = module.exports = DynamicRoutesService;
